// https://community.st.com/s/question/0D53W00001K5Tf7SAF/stm32g4-dac-white-noise
// https://community.st.com/s/question/0D53W00002EuyfGSAR/stm32f-dac-not-driving-voltage-to-more-than-0v

#include <stdint.h>
#include <stddef.h>
#include <string.h>
#include <stdlib.h>
#include "stm32f4xx.h"


// 'F429 DISCO
// GREEN_LED  PG13
// RED_LED    PG14
#define GREEN_LED_PORT   GPIOG
#define GREEN_LED_PIN    13

#define RED_LED_PORT     GPIOG
#define RED_LED_PIN      14




// ------- defines missing from the device header -- normally I have the complete collection in a separate header, this is just to keep things in one file
#define SHL <<
#define AND &
#define OR  |

#define SUPERGLUE(a, b, c, d) a##b##c##d
#define GLUE(a, b) SUPERGLUE(a, b, , )

#define PIN_SET(pin)   pin##_PORT->BSRR = (1 SHL (pin##_PIN + 0))
#define PIN_CLR(pin)   pin##_PORT->BSRR = (1 SHL (pin##_PIN + 16))
#define PIN_GET(pin)   ((pin##_PORT->IDR AND (1 SHL pin##_PIN)) ? 1 : 0)

  // GPIOx_MODER - 2 bits per pin
#define GPIO_Mode_In                         0x00  // GPIO Input Mode
#define GPIO_Mode_Out                        0x01  // GPIO Output Mode
#define GPIO_Mode_AlternateFunction          0x02  // GPIO Alternate function Mode
#define GPIO_Mode_AF                         GPIO_Mode_AlternateFunction
#define GPIO_Mode_Analog                     0x03  // GPIO Analog Mode

#define GPIO_PullNone                        0x00
#define GPIO_PullUp                          0x01
#define GPIO_PullDown                        0x02

// -------- utils/frameworks
// trivial loopdelay
static void LoopDelay(volatile uint32_t n) {
	while(n > 0) n--;
}

#define DELAY_CONSTANT 10000000

// -------------- turbo clock --------------------
#define HSI_CLOCK_MHZ 16
#define HSE_CLOCK_MHZ  8   // DISCOF429 has a 8MHz XTAL on the mcu
// #define PLL_CLOCK_MHZ 160  // this is a high enough clock, which doesn't violate the LCD's SPI clock max speed (i.e. it's exactly 10MHz as specified)
#define PLL_CLOCK_MHZ 180  // this is the max clock of the mcu; it slightly overclocks the LCD's SPI clock


// #define CLOCK_MHZ  HSE_CLOCK_MHZ
// #define CLOCK_MHZ  HSI_CLOCK_MHZ  // default HSI 16MHz clock
 #define CLOCK_MHZ  PLL_CLOCK_MHZ


static void SetSystemClock(void) {

  RCC->APB1ENR |= 0
    | RCC_APB1ENR_PWREN    // to be able to change VOS and overdrive, needed for high frequency gear
  ;


#if (CLOCK_MHZ == PLL_CLOCK_MHZ)
  // gear up system clock
  RCC->CR |= ((uint32_t)RCC_CR_HSEON);  // enable HSE
  while((RCC->CR AND RCC_CR_HSERDY) == 0);  // wait until up

  // before firing up PLL, one word of caution: the VOS bits in PWR_CR register must be 0b11 for fSYS>144MHz
  // that's the reset value so there's no need to change it here (at lower clock, we could lower VOS for lower consumption)
  // if we'd need to change it, the power module's clock would have to be fired up, too - see the STM-supplied "systemInit" code

  // but we want overdrive (i.e. >168MHz system clock, namely the maximum 180MHz) so we need PWR to have enabled clock anyway
  //    During the Over-drive switch activation, no peripheral clocks should be enabled. The peripheral clocks must be enabled once the Over-drive mode is activated.

  // 1.  Select HSI or HSE as system clock.
  // assuming we are on the default HSI now

  // 2.  Configure RCC_PLLCFGR register and set PLLON bit of RCC_CR register.
  RCC->PLLCFGR =    ( HSE_CLOCK_MHZ / 2  * RCC_PLLCFGR_PLLM_0)     // M = XTAL divider - the PLL needs 1MHz to 2MHz at its input, 2MHz reduces jitter
                 OR ( /* 360 */ (2 * PLL_CLOCK_MHZ) / 2  * RCC_PLLCFGR_PLLN_0)    // N = PLL multiplier (feedback divider) - the VCO's output needs to fall between 64MHz and 432MHz
                 OR (( 2  / 2 - 1) * RCC_PLLCFGR_PLLP_0)      // P = divider for SYSCLK - 0 = /2, 1 = /4, 2 = /6, 3 = /8
                 OR ( 8  * RCC_PLLCFGR_PLLQ_0)      // Q = divider for USB (and SDIO and RNG) -- can't achieve 48MHz with 180MHz system clock anyway, we would need to go down to 168MHz (which wouldn't need overdrive btw.), but we don't use USB here anyway
                 OR (RCC_PLLCFGR_PLLSRC_HSE);     // PLL source is HSE

  RCC->CR |= RCC_CR_PLLON;  // fire up PLL

  // 3.  Set ODEN bit of PWR_CR register to enable the Over-drive mode and wait for the ODRDY flag to be set in the PWR_CSR register.
  PWR->CR |= PWR_CR_ODEN;
  while((PWR->CSR AND PWR_CSR_ODRDY) == 0);
  // 4.  Set the ODSW bit in the PWR_CR register to switch the voltage regulator from Normal mode to Over-drive mode. The System will be stalled during the switch but the PLL clock system will be still running during locking phase.
  PWR->CR |= PWR_CR_ODSWEN;
  // 5.  Wait for the ODSWRDY flag in the PWR_CSR to be set.
  while((PWR->CSR AND PWR_CSR_ODSWRDY) == 0);

  // 6.  Select the required Flash latency as well as AHB and APB prescalers.
  FLASH->ACR =  0
    OR FLASH_ACR_ICEN          // meantime, configure flash for maximum performance - enable both caches
    OR FLASH_ACR_DCEN
    OR FLASH_ACR_PRFTEN        // and enable prefetch
    OR FLASH_ACR_LATENCY_5WS   // for VCC>2.7V and FSYS>150MHz, 5 waitstate is appropriate
  ;
  RCC->CFGR = (RCC->CFGR AND (~(0
    OR RCC_CFGR_PPRE1
    OR RCC_CFGR_PPRE2
  ))) OR (0
    OR RCC_CFGR_PPRE1_DIV4   // APB1 prescaler set to 4 -> APB1 clock = 45MHz (required to be <= 45MHz)
    OR RCC_CFGR_PPRE2_DIV2   // APB2 prescaler set to 2 -> APB2 clock = 90MHz (required to be <= 90MHz)
  );
  // 7.  Wait for PLL lock.
  while((RCC->CR AND RCC_CR_PLLRDY) == 0);

  // 8.  Switch the system clock to the PLL.
  RCC->CFGR = (RCC->CFGR AND (~(RCC_CFGR_SW))) OR RCC_CFGR_SW_PLL;  // switch clock source to PLL
  while ((RCC->CFGR AND RCC_CFGR_SWS ) != RCC_CFGR_SWS_PLL);  // wait until switches

  // and that's it.
#elif (CLOCK_MHZ == HSE_CLOCK_MHZ)
  RCC->CR |= ((uint32_t)RCC_CR_HSEON);  // enable HSE
  while((RCC->CR AND RCC_CR_HSERDY) == 0);  // wait until up
  // Switch the system clock to the HSE
  RCC->CFGR = (RCC->CFGR AND (~(RCC_CFGR_SW))) OR RCC_CFGR_SW_HSE;  // switch clock source to PLL
  while ((RCC->CFGR AND RCC_CFGR_SWS ) != RCC_CFGR_SWS_HSE);  // wait until switches
#else // (CLOCK_MHZ == HSI_CLOCK_MHZ)
  // default, nothing to do
#endif
}


// ---------------------------- MAIN ------------------------------
int main(void) {
  SetSystemClock();

  RCC->AHB1ENR |= 0
    | RCC_AHB1ENR_GPIOAEN   // let's enable all GPIOs, we are going to use them all anyway, and don't plan to spare power significantly, so what
    | RCC_AHB1ENR_GPIOBEN
    | RCC_AHB1ENR_GPIOCEN
    | RCC_AHB1ENR_GPIODEN
    | RCC_AHB1ENR_GPIOEEN
    | RCC_AHB1ENR_GPIOFEN
    | RCC_AHB1ENR_GPIOGEN
    | RCC_AHB1ENR_GPIOHEN
  ;
  RCC->APB1ENR |= 0
    | RCC_APB1ENR_TIM6EN
    | RCC_APB1ENR_DACEN
  ;

  GPIOG->MODER = (GPIOG->MODER
  ) | (0
    | (GPIO_Mode_Out  * GPIO_MODER_MODER13_0)   // GREEN_LED
    | (GPIO_Mode_Out  * GPIO_MODER_MODER14_0)   // RED_LED
  );
  // LEDs don't need to change OSPEEDR nor any other GPIO register
  GPIOG->OSPEEDR = 0
    | (2 SHL GPIO_OSPEEDR_OSPEED13_Pos)
  ;


  // set up TIM6->DAC
  DAC->CR = 0
    OR (0 SHL DAC_CR_TSEL2_Pos) // 0 -- use TIM6_TRGO as trigger
    OR DAC_CR_TEN2  // use trigger

#if (1)  // generate triangle
    OR (0b10 SHL DAC_CR_WAVE2_Pos)    // 0b10 - triangle wave generation
#else // generate noise
    OR (0b01 SHL DAC_CR_WAVE2_Pos)    // 0b10 - noise wave generation
#endif
    OR (0b1011 SHL DAC_CR_MAMP2_Pos)  // 0b1011 -- full noise/triangle amplitude

    OR DAC_CR_EN2   // enable
  ;

  TIM6->ARR = 20;  // 42MHz/(20+1) -> cca 2MHz
  TIM6->CR2 = 0
    OR (0b010 SHL TIM_CR2_MMS_Pos)  // 010 Update - The update event is selected as a trigger output (TRGO).
  ;
  TIM6->CR1 = 0
    OR TIM_CR1_CEN  // enable
  ;



  // main loop
  while(1) {
    PIN_SET(RED_LED);
    LoopDelay(DELAY_CONSTANT);
    PIN_CLR(RED_LED);
    LoopDelay(DELAY_CONSTANT);
    __asm("nop");
  }
}



